/*

   Quasar's Bin2C

   Because other Bin2C's aren't good enough! (TM)

   12/25/2009

   Released under the GNU General Public License.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef unsigned char byte;

byte *inputbuffer;
long inputsize;

//
// FatalError
//
static void FatalError(const char *msg)
{
   puts(msg);
   exit(-1);
}

//
// GetInputSize
//
static long GetInputSize(FILE *infile)
{
   long currentPos;
   long fileSize;

   currentPos = ftell(infile);
   fseek(infile, 0, SEEK_END);
   fileSize = ftell(infile);
   fseek(infile, currentPos, SEEK_SET );
   
   return fileSize;
}

//
// ReadInput
//
static void ReadInput(const char *filename)
{
   FILE *infile;

   if(!(infile = fopen(filename, "rb")))
      FatalError("Cannot open input file.");

   inputsize = GetInputSize(infile);

   // allocate input buffer
   if(!(inputbuffer = malloc((size_t)inputsize)))
      FatalError("Could not allocate input buffer.");

   // read input
   fread(inputbuffer, 1, inputsize, infile);

   // done with the input file
   fclose(infile);
}

//
// WriteOutput
//
static void WriteOutput(const char *inputname, const char *outputname, 
                        const char *symbolname)
{
   FILE *outfile;
   int i;

   // open output file
   if(!(outfile = fopen(outputname, "w")))
      FatalError("Cannot open designated output file for writing.");

   // write a header comment
   fprintf(outfile, "// %s\n", outputname);
   fprintf(outfile, "// Generated by qbin2c from %s\n\n", inputname);

   // write opening ifdef guard
   fprintf(outfile, "#ifndef %s_H__\n"
                    "#define %s_H__\n\n", symbolname, symbolname);

   // write array header
   fprintf(outfile, "static unsigned char %s[] =\n{\n", symbolname);

   // write input as C hex bytes
   for(i = 0; i < inputsize; i++)
   {
      fprintf(outfile, " 0x%02x", inputbuffer[i]);
      if(i != inputsize - 1)
         fputs(",", outfile);
      if(i && (!((i + 1) % 12) || i == inputsize - 1))
         fputs("\n", outfile);
   }

   // write closing brace and semicolon
   fputs("};\n\n", outfile);

   // write ifndef closer and EOF comment
   fputs("#endif\n\n// EOF\n\n", outfile);

   // close outfile
   fclose(outfile);
}

//
// Main Routine
//
int main(int argc, const char *argv[])
{
   if(argc < 4)
      FatalError("Usage: qbin2c <infile> <outfile> <symbol>");

   // read input file
   ReadInput(argv[1]);

   // write output file
   WriteOutput(argv[1], argv[2], argv[3]);

   printf("%s written successfully.\n", argv[2]);

	return 0;
}
